--[[
Name: uiQuery 1.0
Revision: $Rev: 25921 $
Website: http://www.wowace.com/
Description: ui wrapper
Dependencies: AceLibrary
]]

local MAJOR_VERSION = "uiQuery-1.0"
local MINOR_VERSION = "$Revision: 25921 $"

if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary.") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

local prototypes = {}
local events = {"OnAnimFinished", "OnChar", "OnClick", "OnCursorChanged", "OnDoubleClick", "OnDragStart", "OnDragStop", "OnEditFocusGained", "OnEditFocusLost", "OnEnter", "OnEnterPressed", "OnEscapePressed", "OnEvent", "OnHide", "OnHorizontalScroll", "OnHyperlinkClick", "OnHyperlinkEnter", "OnHyperlinkLeave", "OnInputLanguageChanged", "OnKeyDown", "OnKeyUp", "OnLeave", "OnMessageScrollChanged", "OnMouseDown", "OnMouseUp", "OnMouseWheel", "OnReceiveDrag", "OnScrollRangeChanged", "OnShow", "OnSizeChanged", "OnSpacePressed", "OnTabPressed", "OnTextChanged", "OnTextSet", "OnTooltipAddMoney", "OnTooltipCleared", "OnTooltipSetDefaultAnchor", "", "OnTooltipSetItem", "OnTooltipSetSpell", "OnTooltipSetUnit", "OnUpdate", "OnUpdateModel", "OnVerticalScroll", "OnValueChanged",}
prototypes["Frame"] = {
    Size = function(self, h, w, s)
        for k, frame in pairs(self.__frames) do
            if not h and not w and not s then
                return frame:GetHeight(), frame:GetWidth(), frame:GetScale()
            else
                if h then frame:SetHeight(h) end
                if w then frame:SetWidth(w) end
                if s then frame:SetScale(s) end
            end
        end
        return self
    end,
}

prototypes["Region"] = {
    Center = function(self, relTo, offX, offY)
        for k, frame in pairs(self.__frames) do
            frame:SetPoint("CENTER", relTo, "CENTER", offX, offY)
        end
        return self
    end,
}

local eventfuncs = {}

for k, v in pairs(events) do
    local script = v
    eventfuncs[script] = function (self, func) 
        if func or (type(func) == "boolean" and func == false) then
            for k, frame in pairs(self.__frames) do
                if frame:HasScript(script) then frame:SetScript(script, func or nil) end
            end
        elseif #self.__frames == 1 and self.__frames[1]:HasScript(script) then
            return self.__frames[1]:GetScript(script)
        end
        return self
    end
end

local function __handler(self, ...)
    if select('#', ...) > 0 then return ... else return self end
end

local function __exp(self, exp, scope, scopes)
    local frames, method, exp2 = {}, "GetChildren"
    
    if exp:match(">") then 
        exp, _, exp2 = exp:match("(%w+) *(.) *(.+)") -- middle return value is for future use.
    end
    
    if exp == "FontString" or exp == "Texture" then method = "GetRegions" end
    if scope and scopes then error("You may only have one type of scope or the other", 2) end
    
    if scope then
        for k, obj in pairs(scope[method] and {scope[method](scope)} or scope) do
            if obj:IsObjectType(exp) then table.insert(frames, obj) end
        end
    elseif scopes then
        for k, scope in pairs(scopes) do
            for k, obj in pairs(scope[method] and {scope[method](scope)} or scope) do
                if obj:IsObjectType(exp) then table.insert(frames, obj) end
            end
        end
    end
    
    if exp2 and #frames > 0 then return __exp(self, exp2, nil, frames) end
    
    return frames
end

local function __index(self, key)
    local key = key
    
    self[key] = function(self, ...)
        if #self.__frames > 1 then 
            for k, frame in pairs(self.__frames) do
                if frame[key] then __handler(self, frame[key](frame, ...)) end
            end
            return self
        else
            if self.__frames[1][key] then return __handler(self, self.__frames[1][key](self.__frames[1], ...)) end
        end
    end
    
    return self[key]
end

local uiQuery = {}

function uiQuery:Create(type, name, parent, template)
    return self:__call(CreateFrame(type, name, parent, template))
end

function uiQuery:__call(...)
    if select('#', ...) == 0 then return self end
    local arg1 = ...
    local frames
    
    if type(arg1) == "table" then
        frames = {...}
    elseif arg1:match("!") then
        local arg2, arg3, arg4 = select(2, ...)
        arg1 = arg1:match("!(%w+)")
        frames = {CreateFrame(arg1, arg2, arg3, arg4)}
    else
        local exp, scope = ...
        if not scope or type(scope) ~= "table" or type(exp) ~= "string" or not scope:IsObjectType("Frame") then
            error("You must pass in a valid expression and scope to use this form of the constructor", 2)
        end
        
        frames = __exp(self, exp, scope)
    end
    
    if not self.registry[frames] then
        local query = { __frames = frames, __index = __index}
        
        for key, value in pairs(prototypes) do
            if frames[1]:IsObjectType(key) then
                for k2, v2 in pairs(value) do
                    if not query[k2] then query[k2] = v2 end
                end
            end
        end
        
        for key, value in pairs(eventfuncs) do
            query[key] = value
        end
        
        setmetatable(query, query)
        self.registry[frames] = query
    end
    
    return self.registry[frames]
end

local function activate(self, oldLib, oldDeactivate)
    uiQuery = self
    
    if oldLib then
        self.registry = oldLib.registry
    else
        self.registry = {}
    end
    
    setmetatable(uiQuery, uiQuery)
    
    if oldDeactivate then
        oldDeactivate(oldLib)
    end
end

AceLibrary:Register(uiQuery, MAJOR_VERSION, MINOR_VERSION, activate)
uiQuery = AceLibrary(MAJOR_VERSION)